import miney


class Inventory:
    """
    Inventories are places to store items, like Chests or player inventories.
    """

    def __init__(self, minetest: miney.Minetest, parent: object):
        self.mt = minetest
        self.parent = parent

    def add(self, item: str, amount: int = 1) -> None:
        """
        Add an item to an inventory. Possible items can be obtained from :attr:`~miney.Node.type`.

        :param item: item type
        :param amount: item amount
        :return: None
        """
        if isinstance(self.parent, miney.Player):
            self.mt.lua.run(
                f"minetest.get_inventory("
                f"{{type = \"player\", name = \"{self.parent.name}\"}}"
                f"):add_item(\"main\", ItemStack(\"{item} {amount}\"))"
            )

    def remove(self, item: str, amount: int = 1) -> None:
        """
        Remove an item from an inventory. Possible items can be obtained from mt.node.type.

        :param item: item type
        :param amount: item amount
        :return: None
        """
        if isinstance(self.parent, miney.Player):
            self.mt.lua.run(
                f"minetest.get_inventory({{type = \"player\", "
                f"name = \"{self.parent.name}\"}}):remove_item(\"main\", ItemStack(\"{item} {amount}\"))")



from typing import Union
import miney

# todo: set modes creative/survival -> Not possible without installed minetest mods


class Player:
    """
    A player of the minetest server.
    """
    def __init__(self, minetest: miney.Minetest, name):
        """
        Initialize the player object.

        :param minetest: Parent minetest object
        :param name: Player name
        """
        self.mt = minetest
        self.name = name
        
        # get user data: password hash, last login, privileges
        data = self.mt.lua.run("return minetest.get_auth_handler().get_auth('{}')".format(self.name))
        if data and all(k in data for k in ("password", "last_login", "privileges")):  # if we have all keys
            self.password = data["password"]
            self.last_login = data["last_login"]
            self.privileges = data["privileges"]
        else:
            raise miney.PlayerInvalid("There is no player with that name")

        self.inventory: miney.Inventory = miney.Inventory(minetest, self)
        """Manipulate player's inventory.
        
        :Example to add 99 dirt to player "IloveDirt"'s inventory:
        
            >>> import miney
            >>> mt = miney.Minetest()
            >>> mt.player.IloveDirt.inventory.add(mt.node.type.default.dirt, 99)      
            
        :Example to remove 99 dirt from player "IhateDirt"'s inventory:
        
            >>> import miney
            >>> mt = miney.Minetest()
            >>> mt.player.IhateDirt.inventory.remove(mt.node.type.default.dirt, 99)
            
        """

    def __repr__(self):
        return '<minetest player "{}">'.format(self.name)

    @property
    def is_online(self) -> bool:
        """
        Returns the online status of this player.

        :return: True or False
        """
        # TODO: Better check without provoke a lua error
        try:
            if self.name == self.mt.lua.run(
                    "return minetest.get_player_by_name('{}'):get_player_name()".format(self.name)):
                return True
        except miney.LuaError:
            return False

    @property
    def position(self) -> dict:
        """
        Get the players current position.

        :return: A dict with x,y,z keys: {"x": 0, "y":1, "z":2}
        """
        try:
            return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_pos()".format(self.name))
        except miney.LuaError:
            raise miney.PlayerOffline("The player has no position, he could be offline")

    @position.setter
    def position(self, values: dict):
        """
        Set player position
        :param values:
        :return:
        """
        if all(k in values for k in ("x", "y", "z")):  # if we have all keys
            self.mt.lua.run(
                "return minetest.get_player_by_name('{}'):set_pos({{x = {}, y = {}, z = {}}})".format(
                    self.name,
                    values["x"],
                    values["y"],
                    values["z"],
                )
            )
        else:
            raise miney.NoValidPosition(
                "A valid position need x,y,z values in an dict ({\"x\": 12, \"y\": 70, \"z\": 12}).")

    @property
    def speed(self) -> int:
        """
        Get or set the players speed. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["speed"]

    @speed.setter
    def speed(self, value: int):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{speed = {}}})".format(self.name, value))

    @property
    def jump(self):
        """
        Get or set the players jump height. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["jump"]

    @jump.setter
    def jump(self, value):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{jump = {}}})".format(self.name, value))

    @property
    def gravity(self):
        """
        Get or set the players gravity. Default is 1.

        :return: Float
        """
        return self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):get_physics_override()".format(self.name))["gravity"]

    @gravity.setter
    def gravity(self, value):
        self.mt.lua.run(
            "return minetest.get_player_by_name('{}'):set_physics_override({{gravity = {}}})".format(self.name, value))

    @property
    def look(self) -> dict:
        """
        Get and set look in radians. Horizontal angle is counter-clockwise from the +z direction. Vertical angle ranges
        between -pi/2 (~-1.563) and pi/2 (~1.563), which are straight up and down respectively.

        :return: A dict like {'v': 0.34, 'h': 2.50} where h is horizontal and v = vertical
        """

        return self.mt.lua.run(
            f"return {{"
            f"h=minetest.get_player_by_name('{self.name}'):get_look_horizontal(), "
            f"v=minetest.get_player_by_name('{self.name}'):get_look_vertical()"
            f"}}"
        )

    @look.setter
    def look(self, value: dict):
        if type(value) is dict:
            if "v" in value and "h" in value:
                if type(value["v"]) in [int, float] and type(value["h"]) in [int, float]:
                    self.mt.lua.run(
                        f"""
                        local player = minetest.get_player_by_name('{self.name}')
                        player:set_look_horizontal({value["h"]})
                        player:set_look_vertical({value["v"]})
                        return true
                        """
                    )
                else:
                    raise TypeError("values for v or h aren't float or int")
            else:
                raise TypeError("There isn't the required v or h key in the dict")
        else:
            raise TypeError("The value isn't a dict, as required. Use a dict in the form: {\"h\": 1.1, \"v\": 1.1}")

    @property
    def look_vertical(self):
        """
        Get and set pitch in radians. Angle ranges between -pi/2 (~-1.563) and pi/2 (~1.563), which are straight
        up and down respectively.

        :return: Pitch in radians
        """
        return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_look_vertical()".format(self.name))

    @look_vertical.setter
    def look_vertical(self, value):
        self.mt.lua.run("return minetest.get_player_by_name('{}'):set_look_vertical({})".format(self.name, value))

    @property
    def look_horizontal(self):
        """
        Get and set yaw in radians. Angle is counter-clockwise from the +z direction.

        :return: Pitch in radians
        """
        return self.mt.lua.run("return minetest.get_player_by_name('{}'):get_look_horizontal()".format(self.name))

    @look_horizontal.setter
    def look_horizontal(self, value):
        self.mt.lua.run("return minetest.get_player_by_name('{}'):set_look_horizontal({})".format(self.name, value))

    @property
    def hp(self):
        """
        Get and set the number of hitpoints (2 * number of hearts) between 0 and 20.
        By setting his hitpoint to zero you instantly kill this player.

        :return:
        """
        return self.mt.lua.run(f"return minetest.get_player_by_name('{self.name}'):get_hp()")

    @hp.setter
    def hp(self, value: int):
        if type(value) is int and value in range(0, 21):
            self.mt.lua.run(
                f"return minetest.get_player_by_name('{self.name}'):set_hp({value}, {{type=\"set_hp\"}})")
        else:
            raise ValueError("HP has to be between 0 and 20.")

    @property
    def breath(self):
        return self.mt.lua.run(f"return minetest.get_player_by_name('{self.name}'):get_breath()")

    @breath.setter
    def breath(self, value: int):
        if type(value) is int and value in range(0, 21):
            self.mt.lua.run(
                f"return minetest.get_player_by_name('{self.name}'):set_breath({value}, {{type=\"set_hp\"}})")
        else:
            raise ValueError("HP has to be between 0 and 20.")

    @property
    def fly(self) -> bool:
        """
        Get and set the privilege to fly to this player. Press K to enable and disable fly mode.
        As a shortcut you can set fly to a number instead if `True` to also changes the players speed to this number.

        .. Example:

            >>> mt.player.MineyPlayer.fly = True  # the can player fly
            >>> mt.player.MineyPlayer.fly = 5  # the can player fly 5 times faster

        :return:
        """
        return self.mt.lua.run(
            f"""
            local privs = minetest.get_player_privs(\"{self.name}\")
            if privs["fly"] then
                return true
            else
                return false
            end
            """
        )

    @fly.setter
    def fly(self, value: Union[bool, int]):
        if value:
            state = "true"
            if type(value) is int:
                if value > 0:
                    self.speed = value
        else:
            state = "false"
        self.mt.lua.run(
            f"""
            local privs = minetest.get_player_privs(\"{self.name}\")
            privs["fly"] = {state}
            minetest.set_player_privs(\"{self.name}\", privs)
            """
        )

    @property
    def creative(self) -> bool:
        return self.mt.lua.run(
            f"""
            local privs = minetest.get_player_privs(\"{self.name}\")
            if privs["creative"] then
                return true
            else
                return false
            end
            """
        )

    @creative.setter
    def creative(self, value: bool):
        if type(value) is not bool:
            raise ValueError("creative needs to be true or false")
        if value is True:
            state = "true"
        else:
            state = "false"

        luastring = f"""
            local privs = minetest.get_player_privs(\"{self.name}\")
            privs["creative"] = {state}
            minetest.set_player_privs(\"{self.name}\", privs)
            """
        self.mt.lua.run(
            luastring
        )


class PlayerIterable:
    """Player, implemented as iterable for easy autocomplete in the interactive shell"""
    def __init__(self, minetest: miney.Minetest, online_players: list = None):
        if online_players:
            self.__online_players = online_players
            self.__mt = minetest

            # update list
            for player in online_players:
                self.__setattr__(player, miney.Player(minetest, player))

    def __iter__(self):
        player_object = []
        for player in self.__online_players:
            player_object.append(miney.Player(self.__mt, player))

        return iter(player_object)

    def __getitem__(self, item_key) -> Player:
        if item_key in self.__online_players:
            return self.__getattribute__(item_key)
        else:
            if type(item_key) == int:
                return self.__getattribute__(self.__online_players[item_key])
            raise IndexError("unknown player")

    def __len__(self):
        return len(self.__online_players)


import socket
import json
import math
from typing import Dict, Union
import time
import string
from random import choices
import miney


class Minetest:
    """__init__([server, playername, password, [port]])
    The Minetest server object. All other objects are accessable from here. By creating an object you connect to Minetest.

    **Parameters aren't required, if you run miney and minetest on the same computer.**

    *If you connect over LAN or Internet to a Minetest server with installed mineysocket, you have to provide a valid
    playername with a password:*

    ::

        >>> mt = Minetest("192.168.0.2", "MyNick", "secret_password")

    Account creation is done by starting Minetest and connect to a server with a username
    and password. https://wiki.minetest.net/Getting_Started#Play_Online

    :param str server: IP or DNS name of an minetest server with installed apisocket mod
    :param str playername: A name to identify yourself to the server.
    :param str password: Your password
    :param int port: The apisocket port, defaults to 29999
    """
    def __init__(self, server: str = "127.0.0.1", playername: str = None, password: str = "", port: int = 29999):
        """
        Connect to the minetest server.

        :param server: IP or DNS name of an minetest server with installed apisocket mod
        :param port: The apisocket port, defaults to 29999
        """
        self.server = server
        self.port = port
        if playername:
            self.playername = playername
        else:
            self.playername = miney.default_playername
        self.password = password

        # setup connection
        self.connection = None
        self._connect()

        self.event_queue = []  # List for collected but unprocessed events
        self.result_queue = {}  # List for unprocessed results
        self.callbacks = {}

        self.clientid = None  # The clientid we got from mineysocket after successful authentication
        self._authenticate()

        # objects representing local properties
        self._lua: miney.lua.Lua = miney.Lua(self)
        self._chat: miney.chat.Chat = miney.Chat(self)
        self._node: miney.node.Node = miney.Node(self)

        player = self.lua.run(
            """
            local players = {}
            for _,player in ipairs(minetest.get_connected_players()) do
                table.insert(players,player:get_player_name())
            end
            return players
            """
        )
        if player:
            player = [] if len(player) == 0 else player
        else:
            raise miney.DataError("Received malformed player data.")

        self._player = miney.PlayerIterable(self, player)

        self._tools_cache = self.lua.run(
            """
            local nodes = {}
            for name, def in pairs(minetest.registered_tools) do
                table.insert(nodes, name)
            end return nodes
            """
        )
        self._tool = miney.ToolIterable(self, self._tools_cache)

    def _connect(self):
        # setup connection
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection.settimeout(2.0)
        self.connection.connect((self.server, self.port))

    def _authenticate(self):
        """
        Authenticate to mineysocket.

        :return: None
        """
        # authenticate
        self.send({"playername": self.playername, "password": self.password})
        result = self.receive("auth")

        if result:
            if "auth_ok" not in result:
                raise miney.AuthenticationError("Wrong playername or password")
            else:
                self.clientid = result[1]  # Clientid = <IP>:<Port>
        else:
            raise miney.DataError("Unexpected authentication result.")

    def send(self, data: Dict):
        """
        Send json objects to the miney-socket.

        :param data:
        :return:
        """
        chunk_size = 4096
        raw_data: bytes = str.encode(json.dumps(data) + "\n")

        try:
            if len(raw_data) < chunk_size:
                self.connection.sendall(raw_data)
            else:  # we need to break the message in chunks
                for i in range(0, int(math.ceil((len(raw_data)/chunk_size)))):
                    self.connection.sendall(
                        raw_data[i * chunk_size:chunk_size + (i * chunk_size)]
                    )
                    time.sleep(0.01)  # Give luasocket a chance to read the buffer in time
                    # todo: Protocol change, that every chunked message needs a response before sending the next
        except ConnectionAbortedError:
            self._connect()
            self.send(data)

    def receive(self, result_id: str = None, timeout: float = None) -> Union[str, bool]:
        """
        Receive data and events from minetest.

        With an optional result_id this function waits for a result with that id by call itself until the right result
        was received. **If lua.run() was used this is not necessary, cause miney already takes care of it.**

        With the optional timeout the blocking waiting time for data can be changed.

        :Example to receive and print all events:

            >>> while True:
            >>>     print("Event received:", mt.receive())

        :param str result_id: Wait for this result id
        :param float timeout: Block further execution until data received or timeout in seconds is over.
        :rtype: Union[str, bool]
        :return: Data from mineysocket
        """

        def format_result(result_data):
            if type(result_data["result"]) in (list, dict):
                if len(result_data["result"]) == 0:
                    return
                if len(result_data["result"]) == 1:  # list with single element doesn't needs a list
                    return result_data["result"][0]
                if len(result_data["result"]) > 1:
                    return tuple(result_data["result"])
            else:
                return result_data["result"]

        # Check if we have to return something received earlier
        if result_id in self.result_queue:
            result = format_result(self.result_queue[result_id])
            del self.result_queue[result_id]
            return result
        # Without a result_id we run an event callback
        elif not result_id and len(self.event_queue):
            result = self.event_queue[0]
            del self.event_queue[0]
            self._run_callback(result)
        try:
            # Set a new timeout to prevent long waiting for a timeout
            if timeout:
                self.connection.settimeout(timeout)
    
            try:
                # receive the raw data and try to decode json
                data_buffer = b""
                while "\n" not in data_buffer.decode():
                    data_buffer = data_buffer + self.connection.recv(4096)
                data = json.loads(data_buffer.decode())
            except socket.timeout:
                raise miney.LuaResultTimeout()
    
            # process data
            if "result" in data:
                if result_id:  # do we need a specific result?
                    if data["id"] == result_id:  # we've got the result we wanted
                        return format_result(data)
                # We store this for later processing
                self.result_queue[data["id"]] = data
            elif "error" in data:
                if data["error"] == "authentication error":
                    if self.clientid:
                        # maybe a server restart or timeout. We just reauthenticate.
                        self._authenticate()
                        raise miney.SessionReconnected()
                    else:  # the server kicked us
                        raise miney.AuthenticationError("Wrong playername or password")
                else:
                    raise miney.LuaError("Lua-Error: " + data["error"])
            elif "event" in data:
                self._run_callback(data)
    
            # if we don't got our result we have to receive again
            if result_id:
                self.receive(result_id)

        except ConnectionAbortedError:
            self._connect()
            self.receive(result_id, timeout)

    def on_event(self, name: str, callback: callable) -> None:
        """
        Sets a callback function for specific events.

        :param name: The name of the event
        :param callback: A callback function
        :return: None
        """
        # Match answer to request
        result_id = ''.join(choices(string.ascii_lowercase + string.ascii_uppercase + string.digits, k=6))
        self.callbacks[name] = callback
        self.send({'activate_event': {'event': name}, 'id': result_id})

    def _run_callback(self, data: dict):
        if data['event'] in self.callbacks:
            # self.callbacks[data['event']](**data['event']['params'])
            if type(data['params']) is dict:
                self.callbacks[data['event']](**data['params'])
            elif type(data['params']) is list:
                self.callbacks[data['event']](*data['params'])

    @property
    def chat(self):
        """
        Object with chat functions.

        :Example:

            >>> mt.chat.send_to_all("My chat message")

        :return: :class:`miney.Chat`: chat object
        """
        return self._chat

    @property
    def node(self):
        """
        Manipulate and get information's about nodes.

        :return: :class:`miney.Node`: Node manipulation functions
        """
        return self._node

    def log(self, line: str):
        """
        Write a line in the servers logfile.

        :param line: The log line
        :return: None
        """
        return self.lua.run('minetest.log("action", "{}")'.format(line))

    @property
    def player(self):
        """
        Get a single players object.

        :Examples:

        Make a player 5 times faster:

            >>> mt.player.MyPlayername.speed = 5

        Use a playername from a variable:

            >>> player = "MyPlayername"
            >>> mt.player[player].speed = 5

        Get a list of all players

            >>> list(mt.player)
            [<minetest player "MineyPlayer">, <minetest player "SecondPlayer">, ...]

        :return: :class:`miney.Player`: Player related functions
        """
        return self._player

    @property
    def lua(self):
        """
        Functions to run Lua inside Minetest.

        :return: :class:`miney.Lua`: Lua related functions
        """
        return self._lua

    @property
    def time_of_day(self) -> int:
        """
        Get and set the time of the day between 0 and 1, where 0 stands for midnight, 0.5 for midday.

        :return: time of day as float.
        """
        return self.lua.run("return minetest.get_timeofday()")

    @time_of_day.setter
    def time_of_day(self, value: float):
        if 0 <= value <= 1:
            self.lua.run("return minetest.set_timeofday({})".format(value))
        else:
            raise ValueError("Time value has to be between 0 and 1.")

    @property
    def settings(self) -> dict:
        """
        Receive all server settings defined in "minetest.conf".

        :return: A dict with all non-default settings.
        """
        return self.lua.run("return minetest.settings:to_table()")

    @property
    def tool(self) -> 'miney.ToolIterable':
        """
        All available tools in the game, sorted by categories. In the end it just returns the corresponding
        minetest tool string, so `mt.tool.default.axe_stone` returns the string 'default:axe_stone'.
        It's a nice shortcut in REPL, cause with auto completion you have only pressed 2-4 keys to get to your type.

        :Examples:

            Directly access a tool:

            >>> mt.tool.default.pick_mese
            'default:pick_mese'

            Iterate over all available types:

            >>> for tool_type in mt.tool:
            >>>     print(tool_type)
            default:shovel_diamond
            default:sword_wood
            default:shovel_wood
            ... (there should be around 34 different tools)
            >>> print(len(mt.tool))
            34

            Get a list of all types:

            >>> list(mt.tool)
            ['default:pine_tree', 'default:dry_grass_5', 'farming:desert_sand_soil', ...

            Add a diamond pick axe to the first player's inventory:

            >>> mt.player[0].inventory.add(mt.tool.default.pick_diamond, 1)

        :rtype: :class:`ToolIterable`
        :return: :class:`ToolIterable` object with categories. Look at the examples above for usage.
        """
        return self._tool

    def __del__(self) -> None:
        """
        Close the connection to the server.

        :return: None
        """
        self.connection.close()

    def __repr__(self):
        return '<minetest server "{}:{}">'.format(self.server, self.port)

    def __delete__(self, instance):
        self.connection.close()



import socket
import time
import os
import platform
import subprocess
import webbrowser
import miney


def is_miney_available(ip: str = "127.0.0.1", port: int = 29999, timeout: int = 1.0) -> bool:
    """
    Check if there is a running miney game available on an optional given host and/or port.
    This functions pings mineysocket and waits **timeout** seconds for a pong.

    :param ip: Optional IP or hostname
    :param port: Optional port
    :param timeout: Optional timeout
    :return: True or False
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(timeout)
    try:
        s.connect((ip, int(port)))
        s.send(b"ping\n")
        reply = s.recv(4096).decode()
        if reply == "pong\n":
            return True
        else:
            return False
    except (socket.timeout, ConnectionResetError, ConnectionRefusedError):
        return False
    finally:
        s.close()


def run_miney_game():
    """
    Run minetest with the miney world. Miney will look for the minetest executable in common places for itself,
    but it's also possible to provide the path as parameter or as environment variable :envvar:`MINETEST_BIN`.

    :return: None
    """
    if is_miney_available():
        raise miney.MinetestRunError("A miney game is already running")
    else:
        run_minetest(show_menu=False)
        wait = 0
        while wait < 12:
            time.sleep(1)
            available = is_miney_available()
            if available:
                time.sleep(2)  # some extra time to get everything initialized
                return True
            wait = wait + 1
        raise miney.MinetestRunError("Timeout while waiting for minetest with an open mineysocket")


def run_minetest(
        minetest_path: str = None,
        show_menu: bool = True,
        world_path: str = "Miney",
        seed: str = "746036489947438842"
) -> None:
    """
    Run minetest. Miney will look for the minetest executable in common places for itself,
    but it's also possible to provide the path as parameter or as environment variable 'MINETEST_BIN'.

    :param minetest_path: Path to the minetest executable
    :param show_menu: Start in the world or in the menu
    :param world_path: Optional world path
    :param seed: Optional world seed
    :return: None
    """
    if not minetest_path:
        if os.environ.get('MINETEST_BIN') and os.path.isfile(os.environ.get('MINETEST_BIN')):
            minetest_path = os.environ['MINETEST_BIN']
        else:
            if platform.system() == 'Windows':
                exe_name = "minetest.exe"
            else:
                exe_name = "minetest"

            # we have to guess the path
            possible_paths = [
                os.path.join(os.getcwd(), "Minetest", "bin"),
            ]
            for p in possible_paths:
                path = os.path.join(p, exe_name)
                if os.path.isfile(path):
                    minetest_path = os.path.join(p, exe_name)
                    break
                else:
                    raise miney.MinetestRunError("Minetest was not found")
    if world_path == "Miney":
        world_path = os.path.abspath(os.path.join(minetest_path, "..", "..", "worlds", "miney"))

        if not os.path.isdir(world_path):  # We have to create the default world
            if not os.path.isdir(os.path.abspath(os.path.join(world_path, "..", "..", "worlds"))):
                os.mkdir(os.path.abspath(os.path.join(world_path, "..", "..", "worlds")))
            os.mkdir(world_path)
            with open(os.path.join(world_path, "world.mt"), "w") as world_config_file:
                world_config_file.write(
                    "enable_damage = true\ncreative_mode = false\ngameid = minetest\nplayer_backend = sqlite3\n"
                    "backend = sqlite3\nauth_backend = sqlite3\nload_mod_mineysocket = true\nserver_announce = false\n"
                )
            with open(os.path.join(world_path, "map_meta.txt"), "w") as world_meta_file:
                world_meta_file.write(f"seed = {seed}")

    if not os.path.isdir(os.path.abspath(os.path.join(minetest_path, "..", "..", "mods", "mineysocket"))):
        raise miney.MinetestRunError("Mineysocket mod is not installed")

    # todo: run_minetest - implementation for linux/macos

    if show_menu:
        subprocess.Popen(
            f"{minetest_path} "
            f"--world \"{world_path}\" --name {miney.default_playername} --address \"\""
        )
    else:
        subprocess.Popen(
            f"{minetest_path} "
            f"--go --world \"{world_path}\" --name {miney.default_playername} --address \"\""
        )


def doc() -> None:
    """
    Open the documention in the webbrower. This is just a shortcut for IDLE or the python interactive console.

    :return: None
    """
    webbrowser.open("https://miney.readthedocs.io/en/latest/")


import miney
from typing import Union
from copy import deepcopy


class Node:
    """
    Manipulate and get information's about nodes.

    **Node manipulation is currently tested for up to 25.000 nodes, more optimization will come later**

    """
    def __init__(self, mt: miney.Minetest):
        self.mt = mt

        self._types_cache = self.mt.lua.run(
            """
            local nodes = {}
            for name, def in pairs(minetest.registered_nodes) do
                table.insert(nodes, name)
            end return nodes
            """
        )
        self._types = TypeIterable(self, self._types_cache)

    @property
    def type(self) -> 'TypeIterable':
        """
        All available node types in the game, sorted by categories. In the end it just returns the corresponding
        minetest type string, so `mt.node.types.default.dirt` returns the string 'default:dirt'.
        It's a nice shortcut in REPL, cause with auto completion you have only pressed 2-4 keys to get to your type.

        :Examples:

            Directly access a type:

            >>> mt.node.type.default.dirt
            'default:dirt'

            Iterate over all available types:

            >>> for node_type in mt.node.type:
            >>>     print(node_type)
            default:pine_tree
            default:dry_grass_5
            farming:desert_sand_soil
            ... (there should be over 400 different types)
            >>> print(len(mt.node.type))
            421

            Get a list of all types:

            >>> list(mt.node.type)
            ['default:pine_tree', 'default:dry_grass_5', 'farming:desert_sand_soil', ...

            Add 99 dirt to player "IloveDirt"'s inventory:

            >>> mt.player.IloveDirt.inventory.add(mt.node.type.default.dirt, 99)

        :rtype: :class:`TypeIterable`
        :return: :class:`TypeIterable` object with categories. Look at the examples above for usage.
        """
        return self._types

    def set(self, nodes: Union[dict, list], name: str = None, offset: dict = None) -> None:
        """
        Set a single or multiple nodes at given position to another node type
        (something like mt.node.type.default.apple).
        You can get a list of all available nodes with :attr:`~miney.Minetest.node.type`

        A node is defined as a dict with these keys:
         * "x", "y", and "z" keys to define the absolute position
         * "name" for a the node type like "default:dirt" (you can also get that from mt.node.type.default.dirt).
           Dicts without name will be set as "air"
         * some other optional minetest parameters

        **The nodes parameter can be a single dict with the above parameters
        or a list of these dicts for bulk spawning.**

        :Examples:

            Set a single node over :

            >>> mt.node.set(mt.player[0].nodes, mt.node)

        :param nodes: A dict or a list of dicts with node definitions
        :param name: a type name like "default:dirt" as string or from :attr:`~miney.Minetest.node.type`. This overrides
        node names defined in the :attr:`nodes` dict
        :param offset: A dict with "x", "y", "z" keys. All node positions will be added with this values.
        """

        _nodes = deepcopy(nodes)

        if offset:
            if not all(pos in ['x', 'y', 'z'] for pos in offset):
                raise ValueError("offset parameter dict needs y, x and z keys")

        # Set a single node
        if type(_nodes) is dict:
            if offset:
                _nodes["x"] = _nodes["x"] + offset["x"]
                _nodes["y"] = _nodes["y"] + offset["y"]
                _nodes["z"] = _nodes["z"] + offset["z"]
            if name:
                _nodes["name"] = name
            self.mt.lua.run(f"minetest.set_node({self.mt.lua.dumps(_nodes)}, {{name=\"{_nodes['name']}\"}})")

        # Set many blocks
        elif type(_nodes) is list:

            lua = ""
            # Loop over nodes, modify name/type, position/offset and generate lua code
            for node in _nodes:
                # default name to 'air'
                if "name" not in node and not name:
                    node["name"] = "air"

                if name:
                    if "name" not in node or (node["name"] != "air" and node["name"] != "ignore"):
                        node["name"] = name

                if offset:
                    node["x"] = node["x"] + offset["x"]
                    node["y"] = node["y"] + offset["y"]
                    node["z"] = node["z"] + offset["z"]

                if node["name"] != "ignore":
                    lua = lua + f"minetest.set_node(" \
                                f"{self.mt.lua.dumps({'x': node['x'], 'y': node['y'], 'z': node['z']})}, " \
                                f"{{name=\"{node['name']}\"}})\n"
            self.mt.lua.run(lua)

    def get(self, position: dict, position2: dict = None, relative: bool = True,
            offset: dict = None) -> Union[dict, list]:
        """
        Get the node at given position. It returns a dict with the node definition.
        This contains the "x", "y", "z", "param1", "param2" and "name" keys, where "name" is the node type like
        "default:dirt".

        If also position2 is given, this function returns a list of dicts with node definitions. This list contains a
        cuboid of definitions with the diagonal between position and position2.

        You can get a list of all available node types with :attr:`~miney.Minetest.node.type`.

        :param position: A dict with x,y,z keys
        :param position2: Another point, to get multiple nodes as a list
        :param relative: Return relative or absolute positions
        :param offset: A dict with "x", "y", "z" keys. All node positions will be added with this values.
        :return: The node type on this position
        """
        if type(position) is dict and not position2:  # for a single node
            _position = deepcopy(position)
            if offset:
                _position["x"] = _position["x"] + offset["x"]
                _position["y"] = _position["y"] + offset["y"]
                _position["z"] = _position["z"] + offset["z"]

            node = self.mt.lua.run(f"return minetest.get_node({self.mt.lua.dumps(position)})")
            node["x"] = position["x"]
            node["y"] = position["y"]
            node["z"] = position["z"]
            return node
        elif type(position) is dict and type(position2) is dict:  # Multiple nodes
            _position = deepcopy(position)
            _position2 = deepcopy(position2)

            if offset:
                _position["x"] = _position["x"] + offset["x"]
                _position["y"] = _position["y"] + offset["y"]
                _position["z"] = _position["z"] + offset["z"]
                _position2["x"] = _position2["x"] + offset["x"]
                _position2["y"] = _position2["y"] + offset["y"]
                _position2["z"] = _position2["z"] + offset["z"]

            nodes_relative = """
                node["x"] = x - start_x
                node["y"] = y - start_y
                node["z"] = z - start_z
                """

            nodes_absolute = """
                node["x"] = x
                node["y"] = y
                node["z"] = z
                """

            return self.mt.lua.run(
                f"""
                pos1 = {self.mt.lua.dumps(_position)}
                pos2 = {self.mt.lua.dumps(_position2)}
                minetest.load_area(pos1, pos2)
                nodes = {{}}
                if pos1.x <= pos2.x then start_x = pos1.x end_x = pos2.x else start_x = pos2.x end_x = pos1.x end
                if pos1.y <= pos2.y then start_y = pos1.y end_y = pos2.y else start_y = pos2.y end_y = pos1.y end
                if pos1.z <= pos2.z then start_z = pos1.z end_z = pos2.z else start_z = pos2.z end_z = pos1.z end
                
                for x = start_x, end_x do
                  for y = start_y, end_y do
                    for z = start_z, end_z do
                      node = minetest.get_node({{x = x, y = y, z = z}})
                      {nodes_relative if relative else nodes_absolute}
                      nodes[#nodes+1] = node  -- append node
                    end
                  end
                end
                return nodes""", timeout=180)

    def __repr__(self):
        return '<minetest node functions>'


class TypeIterable:
    """Node type, implemented as iterable for easy autocomplete in the interactive shell"""
    def __init__(self, parent, nodes_types=None):

        self.__parent = parent

        if nodes_types:

            # get type categories list
            type_categories = {}
            for ntype in nodes_types:
                if ":" in ntype:
                    type_categories[ntype.split(":")[0]] = ntype.split(":")[0]
            for tc in dict.fromkeys(type_categories):
                self.__setattr__(tc, TypeIterable(parent))

            # values to categories
            for ntype in nodes_types:
                if ":" in ntype:
                    self.__getattribute__(ntype.split(":")[0]).__setattr__(ntype.split(":")[1], ntype)
                else:
                    self.__setattr__(ntype, ntype)  # for 'air' and 'ignore'

    def __iter__(self):
        # todo: list(mt.node.type.default) should return only default group
        return iter(self.__parent._types_cache)

    def __getitem__(self, item_key):
        if type(self.__parent) is not type(self):  # if we don't have a category below
            return self.__getattribute__(item_key)
        if item_key in self.__parent.node_types:
            return item_key
        else:
            if type(item_key) == int:
                return self.__parent.node_types[item_key]
            raise IndexError("unknown node type")

    def __len__(self):
        return len(self.__parent._types_cache)



import string
from random import choices
import miney


class Callback:
    """Register callbacks inside minetest, to receive events from Lua"""
    def __init__(self, mt: miney.Minetest):
        self.mt = mt

    def activate(self, event: str, callback: callable, parameters: dict = None):
        """
        Register a callback for an event.

        :param event: Event to register for
        :param callback: The function to be called on events
        :param parameters: Some events need parameters
        :return: None
        """
        # Match answer to request
        result_id = ''.join(choices(string.ascii_lowercase + string.ascii_uppercase + string.digits, k=6))

        self.mt.send(
            {
                'activate_event':
                    {
                        'event': event,
                    },
                'id': result_id
            }
        )



"""
Miney is the python interface to minetest
"""
from .minetest import Minetest
from .player import Player
from .chat import Chat
from .node import Node
from .lua import Lua
from .inventory import Inventory
from .exceptions import *
from .tool import ToolIterable
from .player import PlayerIterable
from .helper import *

__version__ = "0.2.2"
default_playername = "MineyPlayer"

